/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.callcontext;

import static org.assertj.core.api.Assertions.*;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.test.context.support.WithAnonymousUser;
import org.springframework.security.test.context.support.WithMockUser;

import de.ozgcloud.common.test.ITCase;

@ITCase
class CurrentUserServiceITCase {

	@Autowired
	private CurrentUserService service;

	@DisplayName("Get user")
	@Nested
	class TestGetUser {

		@Test
		void shouldThrowExceptionIfUserIsMissing() {
			assertThatThrownBy(() -> service.getUser()).isInstanceOf(IllegalStateException.class);
		}

		@Test
		@WithAnonymousUser
		void shouldThrowExceptionIfAnonymouse() {
			assertThatThrownBy(() -> service.getUser()).isInstanceOf(IllegalStateException.class);
		}

		@Test
		@WithMockUser
		void shouldReturnUser() {
			var user = service.getUser();

			assertThat(user.getName()).isEqualTo("user");
		}

		@Test
		void shouldReturnUserFromToken() {
			var inUser = CallContextUserTestFactory.create();
			SecurityContextHolder.getContext().setAuthentication(CallContextAuthenticationToken.authenticate(inUser));

			var user = service.getUser();

			assertThat(user).isSameAs(inUser);
		}
	}

	@DisplayName("Find user")
	@Nested
	class TestFindUser {

		@Test
		void shouldReturnEmptyIfUserIsMissing() {
			var foundUser = service.findUser();

			assertThat(foundUser).isEmpty();
		}

		@Test
		@WithAnonymousUser
		void shouldTReturnEmptyIfAnonymouse() {
			var foundUser = service.findUser();

			assertThat(foundUser).isEmpty();
		}

		@Test
		@WithMockUser
		void shouldReturnUser() {
			var foundUser = service.findUser();

			assertThat(foundUser.get().getName()).isEqualTo("user");
		}

		@Test
		void shouldReturnUserFromToken() {
			var inUser = CallContextUserTestFactory.create();
			SecurityContextHolder.getContext().setAuthentication(CallContextAuthenticationToken.authenticate(inUser));

			var user = service.findUser();

			assertThat(user).contains(inUser);
		}
	}
}