package de.ozgcloud.vorgang.status;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.function.Predicate;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandRevokeFailedEvent;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.command.PersistedCommand;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemService;
import de.ozgcloud.vorgang.command.CommandCreatedEventTestFactory;

class StatusEventListenerTest {

	@Spy
	@InjectMocks
	private StatusEventListener eventListener;

	@Mock
	private StatusService statusService;
	@Mock
	private VorgangService vorgangService;
	@Mock
	private VorgangAttachedItemService attachedItemService;

	@Mock
	private CommandService commandService;
	@Mock
	private CommandCreatedEvent event;
	@Mock
	private Command command;

	@Mock
	private ApplicationEventPublisher publisher;

	@DisplayName("Test handling Vorgang status events")
	@Nested
	class TestVorgangStatusEvents {

		@BeforeEach
		void init() {
			when(event.getSource()).thenReturn(command);
			when(vorgangService.getById(any())).thenReturn(VorgangTestFactory.create());
			when(commandService.findById(any())).thenReturn(Optional.of(CommandTestFactory.create()));
		}

		@Test
		void shouldCallVorgangService() {
			eventListener.abschiessen(event);

			verify(vorgangService).getById(any());
		}

		@Test
		void shouldCallCommandService() {
			eventListener.abschiessen(event);

			verify(commandService).findById(any());
		}

		@Test
		void shouldHandleVorgangAbschliessen() {
			eventListener.abschiessen(event);

			verify(statusService).abschliessen(any());
		}

		@Test
		void shouldHandleVorgangAnnehmen() {
			eventListener.annehmen(event);

			verify(statusService).annehmen(any());
		}

		@Test
		void shouldHandleVorgangBearbeiten() {
			eventListener.bearbeiten(event);

			verify(statusService).bearbeiten(any());
		}

		@Test
		void shouldHandleVorgangBescheiden() {
			eventListener.bescheiden(event);

			verify(statusService).bescheiden(any());
		}

		@Test
		void shouldHandleVorgangVerwerfen() {
			eventListener.verwerfen(event);

			verify(statusService).verwerfen(any());
		}

		@Test
		void shouldHandleVorgangWiedereroeffnen() {
			eventListener.wiedereroeffnen(event);

			verify(statusService).wiedereroeffnen(any());
		}

		@Test
		void shouldHandleVorgangZurueckholen() {
			eventListener.zurueckholen(event);

			verify(statusService).setStatusNeu(any());
		}

		@Test
		void shouldHandleVorgangZurueckstellen() {
			eventListener.zurueckstellen(event);

			verify(statusService).annehmen(any());
		}

		@Test
		void shouldHandleZuLoeschenMarkieren() {
			eventListener.zuLoeschenMarkieren(event);

			verify(statusService).zumLoeschenMarkieren(notNull());
		}
	}

	@DisplayName("Test setting the previous state value")
	@Nested
	class TestPreviousStatus {

		@Test
		void shouldSetPreviousStatusNeu() {
			var map = eventListener.getPreviousStateMap(VorgangTestFactory.create());

			assertThat(map).containsEntry(Vorgang.MONGODB_FIELDNAME_STATUS, Status.NEU);
		}

		@Test
		void shouldSetPreviousStatusAbgeschlossen() {
			var map = eventListener.getPreviousStateMap(VorgangTestFactory.createBuilder().status(Status.ABGESCHLOSSEN).build());

			assertThat(map).containsEntry(Vorgang.MONGODB_FIELDNAME_STATUS, Status.ABGESCHLOSSEN);
		}
	}

	@DisplayName("Test handling status revoked events")
	@Nested
	class TestStatusRevokedEvent {

		@Test
		void shouldHandleStatusRevokedEvent() {
			eventListener.statusRevokedEvent(new RevokeCommandEvent(CommandTestFactory.create()));

			verify(statusService).revokeStatusChange(any(PersistedCommand.class));
		}

		@Test
		void shouldPublishFailedEventOnException() {
			doThrow(new RuntimeException()).when(statusService).revokeStatusChange(any());

			eventListener.statusRevokedEvent(new RevokeCommandEvent(CommandTestFactory.create()));

			verify(publisher).publishEvent(any(CommandRevokeFailedEvent.class));
		}
	}

	@DisplayName("Test handling exceptions")
	@Nested
	class TestExceptionHandling {

		@Test
		void shouldHandleNoSuchElementException() {
			when(commandService.findById(any())).thenReturn(Optional.empty());
			when(vorgangService.getById(any())).thenReturn(VorgangTestFactory.create());

			eventListener.annehmen(CommandCreatedEventTestFactory.create(CommandTestFactory.create()));

			verify(publisher).publishEvent(any(CommandFailedEvent.class));
		}

		@Test
		void shouldHandleNotFoundException() {
			when(vorgangService.getById(any())).thenThrow(NotFoundException.class);

			eventListener.annehmen(CommandCreatedEventTestFactory.create(CommandTestFactory.create()));

			verify(publisher).publishEvent(any(CommandFailedEvent.class));
		}
	}

	@DisplayName("Test predicated for filtering CommandCreatedEvents")
	@Nested
	class TestPredicates {

		@ParameterizedTest
		@MethodSource("knownOrdersAndMatchingPredicated")
		void shouldFilter(Order order, Predicate<Command> predicate) {
			assertThat(predicate.test(CommandTestFactory.createBuilder().order(order.name()).build())).isTrue();
		}

		@ParameterizedTest
		@MethodSource("knownOrdersAndMatchingPredicated")
		void shouldNotFilter(Order order, Predicate<Command> predicate) {
			assertThat(predicate.test(CommandTestFactory.createBuilder().order(Order.ASSIGN_USER.name()).build())).isFalse();
		}

		private static Stream<Arguments> knownOrdersAndMatchingPredicated() {
			return Stream.of(
					Arguments.of(Order.VORGANG_ABSCHLIESSEN, StatusEventListener.IS_ABSCHLIESSEN_EVENT),
					Arguments.of(Order.VORGANG_ANNEHMEN, StatusEventListener.IS_ANNEHMEN_EVENT),
					Arguments.of(Order.VORGANG_BEARBEITEN, StatusEventListener.IS_BEARBEITEN_EVENT),
					Arguments.of(Order.VORGANG_BESCHEIDEN, StatusEventListener.IS_BESCHEIDEN_EVENT),
					Arguments.of(Order.VORGANG_VERWERFEN, StatusEventListener.IS_VERWERFEN_EVENT),
					Arguments.of(Order.VORGANG_WIEDEREROEFFNEN, StatusEventListener.IS_WIEDEREROEFFNEN_EVENT),
					Arguments.of(Order.VORGANG_ZURUECKHOLEN, StatusEventListener.IS_ZURUECKHOLEN_EVENT),
					Arguments.of(Order.VORGANG_ZURUECKSTELLEN, StatusEventListener.IS_ZURUECKSTELLEN_EVENT));
		}
	}
}
