/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.data.domain.Page;

import io.grpc.stub.StreamObserver;

class GrpcVorgangServiceTest {

	private static final int LIMIT = 10;
	private static final int OFFSET = 0;
	private static final long TOTAL = 42L;

	@Spy
	@InjectMocks
	private GrpcVorgangService service;
	@Mock
	private VorgangService vorgangService;
	@Mock
	private VorgangHeaderService headerService;
	@Mock
	private VorgangHeaderMapper vorgangHeaderMapper;

	@Mock
	private EingangMapper eingangMapper;
	@Mock
	private VorgangWithEingangMapper vorgangWithEingangMapper;
	@Mock
	private FindVorgangRequestMapper findVorgangRequestMapper;

	@Mock
	private FilterByMapper filterByMapper;

	@Mock
	private IncomingFileMapper incomingFileMapper;
	@Mock
	private IncomingFileGroupMapper incomingFileGroupMapper;

	@Nested
	class TestCreate {

		@Mock
		private StreamObserver<GrpcCreateVorgangResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcCreateVorgangResponse> responseCaptor;

		private final GrpcCreateVorgangRequest request = GrpcCreateVorgangRequest.newBuilder()
				.setEingang(GrpcEingangTestFactory.create())
				.build();
		private final Eingang eingang = EingangTestFactory.create();

		@BeforeEach
		void mockMapperReturnValues() {
			when(eingangMapper.fromGrpc(any())).thenReturn(eingang);
		}

		@Nested
		class TestStartCreation {

			@BeforeEach
			void mockVorgangService() {
				when(vorgangService.startCreation(any())).thenReturn(VorgangTestFactory.create());
			}

			@Test
			void shouldCallEingangMapper() {
				callStartCreation();

				verify(eingangMapper).fromGrpc(GrpcEingangTestFactory.create());
			}

			@Test
			void shouldReturnMessage() {
				callStartCreation();

				assertThat(responseCaptor.getValue().getMessage()).isEqualTo("ok");
			}

			@Test
			void shouldReturnVorgangId() {
				callStartCreation();

				assertThat(responseCaptor.getValue().getVorgangId()).isEqualTo(VorgangTestFactory.ID);
			}

			private void callStartCreation() {
				service.startCreation(request, responseObserver);

				verify(responseObserver).onNext(responseCaptor.capture());
			}
		}
	}

	@Nested
	class TestFinishCreation {

		@Mock
		private StreamObserver<GrpcFinishCreationResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcFinishCreationResponse> responseCaptor;

		@Test
		void shouldCallFileGroupMapper() {
			callService();

			verify(incomingFileGroupMapper).mapGrpc(GrpcFinishCreationRequestTestFactory.attachment);
		}

		@Test
		void shouldCallFileMapper() {
			callService();

			verify(incomingFileMapper).mapGrpc(GrpcFinishCreationRequestTestFactory.representation);
		}

		@Nested
		class TestCallService {

			private final IncomingFileGroup fileGroup = IncomingFileGroupTestFactory.create();
			private final IncomingFile file = IncomingFileTestFactory.create();

			@BeforeEach()
			void mockMapper() {
				when(incomingFileGroupMapper.mapGrpc(any(GrpcIncomingFileGroup.class))).thenReturn(fileGroup);
				when(incomingFileMapper.mapGrpc(any(GrpcIncomingFile.class))).thenReturn(file);
			}

			@Test
			void shouldCallService() {
				callService();

				verify(vorgangService).finishCreation(VorgangTestFactory.ID, List.of(fileGroup), List.of(file));
			}
		}

		@Test
		void shouldCallOnNextWithResponse() {
			callService();

			verify(responseObserver).onNext(responseCaptor.capture());
			assertThat(responseCaptor.getValue().getMessage()).isEqualTo("ok");
		}

		@Test
		void shouldComplete() {
			callService();

			verify(responseObserver).onCompleted();
		}

		private void callService() {
			service.finishCreation(GrpcFinishCreationRequestTestFactory.create(), responseObserver);
		}
	}

	@Nested
	class TestFindVorgange {

		@Mock
		private StreamObserver<GrpcFindVorgangResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcFindVorgangResponse> responseCaptor;
		@Mock
		private Page<VorgangHeader> page;
		@Captor
		private ArgumentCaptor<VorgangHeader> headerCaptor;
		@Captor
		private ArgumentCaptor<FindVorgangRequest> findVorgangRequestCaptor;

		private final GrpcVorgangHeader grpcVorgangHeader = GrpcVorgangHeader.newBuilder().build();
		private final GrpcFindVorgangRequest request = GrpcFindVorgangRequest.newBuilder().setLimit(LIMIT).setOffset(OFFSET)
				.build();

		@BeforeEach
		void initMockReturnValues() {
			when(page.getContent()).thenReturn(Collections.singletonList(VorgangHeaderTestFactory.create()));
			when(page.getTotalElements()).thenReturn(TOTAL);
			when(headerService.findAll(any(FindVorgangRequest.class))).thenReturn(page);
			when(vorgangHeaderMapper.toGrpcVorgangHeader(any())).thenReturn(grpcVorgangHeader);
			when(findVorgangRequestMapper.fromFindVorgangRequest(any(GrpcFindVorgangRequest.class)))
					.thenReturn(FindVorgangRequestTestFactory.create());
		}

		@Test
		void shouldCallVorgangHeaderService() {
			callFindVorgaenge();

			verify(headerService).findAll(findVorgangRequestCaptor.capture());
		}

		@Test
		void shouldCallVorgangHeaderMapper() {
			callFindVorgaenge();

			verify(vorgangHeaderMapper).toGrpcVorgangHeader(any());
		}

		@Test
		void shouldContainVorgangData() {
			callFindVorgaenge();

			verify(vorgangHeaderMapper).toGrpcVorgangHeader(headerCaptor.capture());
			assertThat(headerCaptor.getValue().getId()).isEqualTo(VorgangTestFactory.ID);
			assertThat(headerCaptor.getValue().getCreatedAt()).isEqualTo(VorgangTestFactory.INITIAL_DATE_STR);
		}

		@Test
		void shouldSetTotal() {
			callFindVorgaenge();

			var response = responseCaptor.getValue();

			assertThat(response.getTotal()).isEqualTo(TOTAL);
		}

		private void callFindVorgaenge() {
			service.findVorgang(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@DisplayName("Find vorgang with eingang")
	@Nested
	class TestFindVorgangWithEingang {

		@Mock
		private StreamObserver<GrpcFindVorgangWithEingangResponse> responseObserver;
		@Captor
		private ArgumentCaptor<GrpcFindVorgangWithEingangResponse> responseCaptor;

		private final FilterCriteria filterCriteria = FilterCriteriaTestFactory.create();
		private final Vorgang vorgang = VorgangTestFactory.create();
		private final GrpcVorgangWithEingang vorgangWithEingang = GrpcVorgangWithEingangTestFactory.create();
		private final GrpcFindVorgangWithEingangRequest request = GrpcFindVorgangWithEingangRequest.newBuilder()
				.setId(VorgangTestFactory.ID)
				.setFilterBy(GrpcFilterByTestFactory.create())
				.build();

		@BeforeEach
		void mock() {
			when(vorgangService.getById(anyString(), any())).thenReturn(vorgang);
			when(vorgangWithEingangMapper.toVorgangWithEingang(any())).thenReturn(vorgangWithEingang);
		}

		@DisplayName("service")
		@Nested
		class TestFindVorgangWithEingangService {

			@Captor
			private ArgumentCaptor<Vorgang> vorgangArgumentCaptor;

			@BeforeEach
			void mock() {
				when(filterByMapper.fromFilterBy(any(GrpcFilterBy.class))).thenReturn(filterCriteria);
			}

			@Test
			void shouldCallHeaderService() {
				callFindVorgangWithEingang();

				verify(vorgangService).getById(VorgangTestFactory.ID, filterCriteria);
			}

			@Test
			void shouldCallFindVorgangRequestMapper() {
				callFindVorgangWithEingang();

				verify(filterByMapper).fromFilterBy(GrpcFilterByTestFactory.create());
			}

			@Test
			void shouldCallVorgangWithEingangMapper() {
				callFindVorgangWithEingang();

				verify(vorgangWithEingangMapper).toVorgangWithEingang(vorgang);
			}

			@Test
			void shouldContainVorgangData() {
				callFindVorgangWithEingang();

				assertThat(responseCaptor.getValue().getVorgangWithEingang()).usingRecursiveComparison()
						.isEqualTo(GrpcVorgangWithEingangTestFactory.create());
			}

		}

		private void callFindVorgangWithEingang() {
			service.findVorgangWithEingang(request, responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}
}